/* main_ermg.cc
 *
 * Copyright (C) 2006 Laboratoire Statistique & Gnome
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/**
 * Fix (part I) to avoid macro redefinition warning messages.
 **/
#undef PACKAGE
#undef VERSION
#include "../tclap-1.2.0/config/config.h"
/**
 * Fix (part II) to avoid macro redefinition warning messages.
 **/
#undef PACKAGE
#undef VERSION
#include <tclap/CmdLine.h>
using namespace TCLAP;

#include <sstream>
using namespace std;


#include <sys/types.h>
#include <unistd.h>
#include <Emd.h>
#include <GraphReader.h>
#include <ModelImprover.h>
#include <ModelDesigner.h>

using namespace ermg;

 
int main(int argc, char ** argv )
{
  CmdLine cmdLine(PACKAGE,' ',VERSION);

  SwitchArg rand_arg("","random","Draw random input classes for [a subset of] the vertices (to use without -Q).",false);
  cmdLine.add(rand_arg);
  ValueArg<int> initnbv_arg("t","initnbv","Number of vertices to consider for a random initialisation (to use with --random, Default: all vertices).",false,0,"integer");
  cmdLine.add(initnbv_arg);  
  ValueArg<string> classinit_arg("i","inputclass","Supplies input classes for [a subset of] the vertices (to use without -Q)",false,"","filename");
  cmdLine.add(classinit_arg);
  SwitchArg symmetrize_arg("y","symmetrize","Symmetrizes the matrix.",false);
  cmdLine.add(symmetrize_arg);
  SwitchArg enable_loop_arg("","loop","Enables loop from a vertex to itself.",false);
  cmdLine.add(enable_loop_arg);
  SwitchArg undirected_arg("","undirected","Specifies that the network is undirected.",false);
  cmdLine.add(undirected_arg);
  SwitchArg silent_arg("","silent","Silent mode.",false);
  cmdLine.add(silent_arg);
  ValueArg<int> improve_nbiter_arg("n","improvenbiter","Number of improve iterations (Default: 5).",false,5,"integer");
  cmdLine.add(improve_nbiter_arg);
  ValueArg<int> kmeans_nbclass_arg("k","kmeansnbclass","Activate the k-means initialization (faster but less accurate) for the given number of classes.",false,0,"integer");
  cmdLine.add(kmeans_nbclass_arg);
  ValueArg<int> kmeans_nbiter_arg("K","kmeansnbiter","Number of K-means iterations (Default: 50).",false,50,"integer");
  cmdLine.add(kmeans_nbiter_arg);
  ValueArg<int> em_nbiter_arg("e","emnbiter","Number of EM iterations (Default: 10).",false,10,"integer");
  cmdLine.add(em_nbiter_arg);
  ValueArg<double> em_eps_arg("E","emeps","Epsilon for EM (Default: 1e-10).",false,1e-10,"float");
  cmdLine.add(em_eps_arg);
  ValueArg<int> fp_nbiter_arg("f","fpnbiter","Number of FP iterations (Default: 50).",false,50,"integer");
  cmdLine.add(fp_nbiter_arg);
  ValueArg<double> fp_eps_arg("F","fpeps","Epsilon for FP (Default: 1e-4).",false,1e-4,"float");
  cmdLine.add(fp_eps_arg);
  ValueArg<string> modelinitdir_arg("I","inputdir","Directory of the input models for qmin<=q<=qmax (associated to the matrix file) to improve.",false,"","dirname");
  cmdLine.add(modelinitdir_arg);
  SwitchArg improve_arg("","improve","Performs improvements by split/merge for unsatisfying likelihoods (and sum up the results into a .likelihoods file).", false);
  cmdLine.add(improve_arg);
  SwitchArg speed_arg("","speed","Speed up the improvements (faster but less accurate).", false);
  cmdLine.add(speed_arg);
  ValueArg<string> outdir_arg("o","outputdir","Directory for output files (network directory by default).",false,"","dirname");
  cmdLine.add(outdir_arg);
  ValueArg<int> maxnbclass_arg("Q","qmax","Maximum number of classes.",false,2,"integer");
  cmdLine.add(maxnbclass_arg);
  ValueArg<int> minnbclass_arg("q","qmin","Minimum number of classes.",false,2,"integer");
  cmdLine.add(minnbclass_arg);
  UnlabeledValueArg<string> ermg_file("m", "Network file ( in \"spm\" format).", true, "", "filename");
  cmdLine.add(ermg_file);
  cmdLine.parse(argc,argv);


  srand((long)getpid() * time(NULL));
  if ((modelinitdir_arg.getValue().size()!=0)&&(improve_arg.getValue()==false)){
    cerr<<"Error: Option -I must be associated to option --improve"<<endl;
    exit(1);
  }
  if ( !((improve_arg.getValue()==true)&&(modelinitdir_arg.getValue().size()>0)) 
       && ((classinit_arg.getValue().size()==0) && (rand_arg.getValue()==false)) )    
    if ((kmeans_nbclass_arg.getValue()>0)&&(kmeans_nbclass_arg.getValue()<maxnbclass_arg.getValue())){
      cerr<<"Error: Bad number of classes for the K-means"<<endl;
      exit(1);
    }


  if ( (minnbclass_arg.getValue()<=maxnbclass_arg.getValue()) 
       && ((classinit_arg.getValue().size()!=0)||(rand_arg.getValue()==true)) ){
    cerr<<"Warning: Option -i or --random must be associated to option -q only, -Q ignored"<<endl;
    maxnbclass_arg.getValue()=minnbclass_arg.getValue();
  }
  else
    if (maxnbclass_arg.getValue()<minnbclass_arg.getValue()){
      maxnbclass_arg.getValue()=minnbclass_arg.getValue();
    }



  // GRAPH
  bool tosym = symmetrize_arg.getValue();
  if (undirected_arg.getValue()){
	tosym = true;
  }
  GraphReader gr(tosym, enable_loop_arg.getValue());
  if (!silent_arg.getValue())
    if (undirected_arg.getValue())
      cout<<"Loading undirected network...\t"<<flush;
    else
      cout<<"Loading directed network...\t"<<flush;

  try {

    gr.loadFromFile( ermg_file.getValue() );
    if (!silent_arg.getValue())
      cout<<"done"<<endl;
    
    Graph *graph=gr.getGraph();
    if (graph->nbVertices()<maxnbclass_arg.getValue()){
      cerr<<"Error: Q max is greater than (then set to) the number of vertices"<<endl;
      exit(0);
    }        

    // INPUT CLASSES
    vector<int> inputclass;
    if (classinit_arg.getValue().size()!=0)      
      inputclass = graph->reorder(classinit_arg.getValue());
    else if (rand_arg.getValue()==true)
      graph->shuffle();


    // MODEL 
    Ermg* model;
    if (undirected_arg.getValue()==false){
      model = new Ermdg( graph->getMatrix(), graph->getTransposedMatrix(),
			 enable_loop_arg.getValue(),
			 kmeans_nbiter_arg.getValue(),
			 kmeans_nbclass_arg.getValue(),
			 minnbclass_arg.getValue(), maxnbclass_arg.getValue() );
    }
    else{	
      model = new Ermg( graph->getMatrix(),
			enable_loop_arg.getValue(),
			kmeans_nbiter_arg.getValue(),
			kmeans_nbclass_arg.getValue(),
			minnbclass_arg.getValue(), maxnbclass_arg.getValue() );
    } 

    // METHOD 
    EmCore * em_method = NULL;
    if (undirected_arg.getValue()==false){
      // EM UNDIRECTED
      em_method = new Emd(  dynamic_cast<Ermdg&>(*model),
			    em_nbiter_arg.getValue(), em_eps_arg.getValue(),
			    fp_nbiter_arg.getValue(), fp_eps_arg.getValue() );
    }
    else{
      // EM DIRECTED
      em_method = new Em(  *model,
			   em_nbiter_arg.getValue(), em_eps_arg.getValue(),
			   fp_nbiter_arg.getValue(), fp_eps_arg.getValue() );
    }    

    
    // ESTIMATION
    if ( (classinit_arg.getValue().size()==0) && (rand_arg.getValue()==false) ){
      ModelImprover modelimp( *model, minnbclass_arg.getValue(), maxnbclass_arg.getValue(), improve_nbiter_arg.getValue(), silent_arg.getValue() );
      if (modelinitdir_arg.getValue().size()==0){
	modelimp.initialLikelihoods( *em_method, kmeans_nbclass_arg.getValue()==0 );
      }
      else{	
	modelimp.loadLikelihoods( ermg_file.getValue(), modelinitdir_arg.getValue() );
      } 
      if (improve_arg.getValue()==true){
	if (speed_arg.getValue()==true)
	  modelimp.improveLikelihoodsSpeed( *em_method );
	else
	  modelimp.improveLikelihoods( *em_method );
      }
      modelimp.outFile( ermg_file.getValue(), graph, outdir_arg.getValue() );
    }
    else{      
      ModelDesigner modeldes( *model, minnbclass_arg.getValue(), silent_arg.getValue() );
      int nbv;
      if (classinit_arg.getValue().size()!=0)
	nbv = inputclass.size();
      else{
	nbv = initnbv_arg.getValue();
	if ((nbv>model->nbVertices()) || (nbv==0))
	  nbv = model->nbVertices();
      }
      if (classinit_arg.getValue().size()!=0)
	modeldes.load(inputclass);
      else
	modeldes.initialRandom(nbv);      
      em_method->setRequiredNumberOfInitializedVertices(nbv);
      modeldes.em( *em_method );
      modeldes.outFile( ermg_file.getValue(), graph, outdir_arg.getValue() );
    }
    delete model;
    delete graph;
    delete em_method;

  } catch (GraphReaderException &gre) {
    cerr << "Exception : " << gre.what() << endl;
  }
}
