/* main_ermgo.cc
 *
 * Copyright (C) 2006 Laboratoire Statistique & Gnome
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/**
 * Fix (part I) to avoid macro redefinition warning messages.
 **/
#undef PACKAGE
#undef VERSION
#include "../tclap-1.2.0/config/config.h"
/**
 * Fix (part II) to avoid macro redefinition warning messages.
 **/
#undef PACKAGE
#undef VERSION
#include <tclap/CmdLine.h>
using namespace TCLAP;

#include <sstream>
using namespace std;


#include <sys/types.h>
#include <unistd.h>
#include <SOCEm.h>
#include <OEm.h>
#include <GraphReader.h>
#include <ModelImprover.h>
#include <ModelDesigner.h>


using namespace ermg;

 
int main(int argc, char ** argv )
{
  CmdLine cmdLine(PACKAGE,' ',VERSION);
  
  SwitchArg enable_loop_arg("","loop","Enables loop from a vertex to itself.",false);
  cmdLine.add(enable_loop_arg);
  SwitchArg silent_arg("","silent","Silent mode.",false);
  cmdLine.add(silent_arg);  
  SwitchArg rand_arg("","random","Draw random input classes for [a subset of] the vertices (to use without -Q).",false);
  cmdLine.add(rand_arg);
  ValueArg<int> initnbv_arg("t","initnbv","Minimal number of vertices to consider for initialisation (Default: (vertices number)/3).",false,0,"integer");
  cmdLine.add(initnbv_arg);
  ValueArg<int> improve_nbiter_arg("n","improvenbiter","Number of improve iterations (Default: 5).",false,5,"integer");
  cmdLine.add(improve_nbiter_arg);
  ValueArg<int> em_nbiter_arg("e","emnbiter","Number of EM iterations (Default: 3).",false,3,"integer");
  cmdLine.add(em_nbiter_arg);
  ValueArg<string> classinit_arg("i","inputclass","Supplies input classes for [a subset of] the vertices (to use without -Q)",false,"","filename");
  cmdLine.add(classinit_arg);
  ValueArg<string> modelinitdir_arg("I","inputdir","Directory of the input models for qmin<=q<=qmax (associated to the matrix file) to improve.",false,"","dirname");
  cmdLine.add(modelinitdir_arg);
  SwitchArg improve_arg("","improve","Performs improvements by split/merge for unsatisfying likelihoods (and sum up the results into a .likelihoods file).", false);
  cmdLine.add(improve_arg);
  SwitchArg speed_arg("","speed","Speed up the improvements.", false);
  cmdLine.add(speed_arg);
  ValueArg<string> outdir_arg("o","outputdir","Directory for output files (network directory by default).",false,"","dirname");
  cmdLine.add(outdir_arg);
  SwitchArg classif_arg("","classif","Use the classification algorithm.",false);
  cmdLine.add(classif_arg); 
  SwitchArg stoch_arg("","stochastic","Use the stochastic classification algorithm.",false);
  cmdLine.add(stoch_arg);
  ValueArg<int> maxnbclass_arg("Q","qmax","Maximum number of classes (optionnal).",false,1,"integer");
  cmdLine.add(maxnbclass_arg);
  ValueArg<int> minnbclass_arg("q","qmin","Minimum number of classes.",false,1,"integer");
  cmdLine.add(minnbclass_arg);
  UnlabeledValueArg<string> ermg_file("m", "Network file (in \"spm\" format).", true, "", "filename");
  cmdLine.add(ermg_file);
  cmdLine.parse(argc,argv);


  srand((long)getpid());
  if ((modelinitdir_arg.getValue().size()!=0)&&(improve_arg.getValue()==false)){
    cerr<<"Error: Option -I must be associated to option --improve"<<endl;
    exit(1);
  } 

  if ( (minnbclass_arg.getValue()<=maxnbclass_arg.getValue()) 
       && ((classinit_arg.getValue().size()!=0)||(rand_arg.getValue()==true)) ){
    cerr<<"Warning: Option -i or --random must be associated to option -q only, -Q ignored"<<endl;
    maxnbclass_arg.getValue()=minnbclass_arg.getValue();
  }
  else
    if (maxnbclass_arg.getValue()<minnbclass_arg.getValue()){
      maxnbclass_arg.getValue()=minnbclass_arg.getValue();
    }


  // GRAPH
  bool tosym = true;
  GraphReader gr(tosym, enable_loop_arg.getValue());
  if (!silent_arg.getValue())
    cout<<"Loading network...\t"<<flush;
  try {

    gr.loadFromFile( ermg_file.getValue() );
    if (!silent_arg.getValue())
    cout<<"done"<<endl;
    
    Graph *graph=gr.getGraph();
    if (graph->nbVertices()<maxnbclass_arg.getValue()){
      cerr<<"Error: Q max is greater than (then set to) the number of vertices"<<endl;
      exit(0);
    }
    
    graph->shuffle();

    // INPUT CLASSES
    vector<int> inputclass;
    if (classinit_arg.getValue().size()!=0)
      inputclass = graph->reorder(classinit_arg.getValue());


    // MODEL 
    Ermg* model;
    model = new Ermg( graph->getMatrix(),
		      enable_loop_arg.getValue(),
		      0,
		      0,
		      minnbclass_arg.getValue(), maxnbclass_arg.getValue() );

    
    // METHOD 
    EmCore * em_method = NULL;
    int nbv;
    if (classinit_arg.getValue().size()!=0)
      nbv = inputclass.size();
    else{
      nbv = initnbv_arg.getValue();
      if (nbv==0)
	nbv = model->nbVertices()/3;
      if (nbv<MININITNBV)
	nbv = MININITNBV;
      if (nbv>model->nbVertices())
	nbv = model->nbVertices();
    }

    if (classif_arg.getValue()==true) 
      // CLASSIFICATION ONLINE EM 
      em_method = new OCEm( *model, nbv, em_nbiter_arg.getValue() );	  
    else 
      if (stoch_arg.getValue()==true) 
	// STOCHASTIC CLASSIFICATION ONLINE EM 
	em_method = new SOCEm( *model, nbv, em_nbiter_arg.getValue() );	  
      else 
	// ONLINE EM 
	em_method = new OEm( *model, nbv, em_nbiter_arg.getValue() );	  
      

    // ESTIMATION
    if ( (classinit_arg.getValue().size()==0) && (rand_arg.getValue()==false) ){
      ModelImprover modelimp( *model, minnbclass_arg.getValue(), maxnbclass_arg.getValue(), improve_nbiter_arg.getValue(), silent_arg.getValue() );
      if (modelinitdir_arg.getValue().size()==0){
	modelimp.initialLikelihoods( *em_method );
      }
      else{	
	modelimp.loadLikelihoods( ermg_file.getValue(), modelinitdir_arg.getValue() );
      } 
      if (improve_arg.getValue()==true){
	if (speed_arg.getValue()==true)
	  modelimp.improveLikelihoodsSpeed( *em_method );
	else
	  modelimp.improveLikelihoods( *em_method );
      }      
      modelimp.outFile( ermg_file.getValue(), graph, outdir_arg.getValue() );
    }
    else{
      ModelDesigner modeldes( *model, minnbclass_arg.getValue(), silent_arg.getValue() );
      if (classinit_arg.getValue().size()!=0)
	modeldes.load(inputclass);
      else
	modeldes.initialRandom(nbv);
      modeldes.em( *em_method );
      modeldes.outFile( ermg_file.getValue(), graph, outdir_arg.getValue() );
    }

    delete model;
    delete graph;
    delete em_method;
    
  } catch (GraphReaderException &gre) {
    cerr << "Exception : " << gre.what() << endl;
  }
}
