#!/usr/bin/perl

# Copyright (C) 2006 Laboratoire Statistique & G^[-b-bnome^[-A-A
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Authors:
# Vincent Miele, Franck Picard
# 01/11/2006

use strict;
if ($#ARGV < 2) {
  print "Usage: perl $0 file_Qx.model nbnodes maxdegree [--undirected]\nOR\n";
  print "Usage: perl $0 file_Qx.model graph.spm maxdegree [--undirected]\n";
  exit(0);
}

my $maxdegree=$ARGV[2];
########
my $nbnodes;
if ($ARGV[1]=~/spm/){
  open(SPM, "<$ARGV[1]") or die "File \"$ARGV[1]\" opening problem\n";
  my %refnodes=();
  my $edg=<SPM>;
  while ( defined($edg) ){
    my @vert = split /\s+/, $edg;
    $refnodes{$vert[0]}++;
    $refnodes{$vert[1]}++;
    $edg=<SPM>;
  }
  $nbnodes=scalar(keys%refnodes);
  close(SPM);
}
else{
  $nbnodes=$ARGV[1];
}

########
open MODEL,"$ARGV[0]" or die "File \"$ARGV[0]\" opening problem\n";
# Q
my $Q;
my $l=<MODEL>;
$l=<MODEL>;
if ($l=~/^(\d+)$/){
  $Q=$1;
}
else{
  die "File \"$ARGV[0]\" format problem\n"
}
# Alpha
$l=<MODEL>;
$l=<MODEL>;
chomp($l);
my @al = split /\s+/, $l;
if (($#al+1)!=$Q){
  die "File \"$ARGV[0]\" format problem\n"
}
# Pi
$l=<MODEL>;
$l=<MODEL>;
my @pi=();
while ($l=~/^\d+/){
  chomp($l);
  my @li=split /\s+/, $l;
  if (($#li+1)!=$Q){
    die "File \"$ARGV[0]\" format problem\n";
  }
  push @pi, @li;
  $l=<MODEL>;
}
close MODEL;


########
if ($ARGV[3]=~/undirected/){
  my @pibar=();
  foreach my $q(0..($Q-1)){
    $pibar[$q]=0;
    foreach my $l(0..($Q-1)){
      $pibar[$q]+=$al[$l]*$pi[$q*$Q+$l];
    }
  }
  print "# DEGREE PROBA\n";
  foreach my $k(0..$maxdegree){
    my $prob=0;
    foreach my $q(0..($Q-1)){
      $prob+=$al[$q]*( exp($k*log(($nbnodes)*$pibar[$q])) * exp(-($nbnodes)*$pibar[$q]) ) / fact_loop_b($k);
    }
    print "$k\t$prob\n";
  }
}
else{ 
  my @pibar=();
  foreach my $q(0..($Q-1)){
    $pibar[$q]=0;
    foreach my $l(0..($Q-1)){
      $pibar[$q]+=$al[$l]*$pi[$q*$Q+$l];
    }
  }
  print "# OUT-DEGREE PROBA\n";
  foreach my $k(0..$maxdegree){
    my $prob=0;
    foreach my $q(0..($Q-1)){
      $prob+=$al[$q]*( exp($k*log(($nbnodes)*$pibar[$q])) * exp(-($nbnodes)*$pibar[$q]) ) / fact_loop_b($k);
    }
    print "$k\t$prob\n";
  }
}

# from http://www.perl.org/tpc/1998/Perl_Language_and_Modules/Efficient%20Perl/handout.html
sub fact_loop_b {		# looping, switch to BigInt	
  use Math::BigInt;
  my $n = shift;
  
  return undef if $n <  0;
  return 1     if $n <= 1;
  
  my $prod = my $k = 1;
  ( $prod, $k ) = (Math::BigInt->new( '6227020800' ), 13)
    if $k >= 13;
  
  $prod *= ++$k while $k < $n;
  
  return $prod;
}
