#! /usr/bin/perl

# Copyright (C) 2008 CNRS
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Authors:
# Vincent Miele
# 15/05/2009

use strict;

use Getopt::Long;
my $help = '';
my $undirected = '';
my $loop = '';
my $verbose;
my $netfile = '';
my $nbnodes = 0;
GetOptions('help|h' => \$help,
	   'undirected' => \$undirected,
	   'loop' => \$loop,
	   'verbose' => \$verbose,
           'nbnodes|n=i' => \$nbnodes);


if (($#ARGV < 1)or($help)or($nbnodes<1)){
  print "Usage:\t$0 [FILE.model FILE.class] [OPTION]\n";
  print "From a MixNet .model file and a .class file, simulate a network eXtension.\n";
  print "\nOptions:\n\t-n <int>\tnumber of nodes to add to extend the network by simulation\n";
  print "\t-i <file>\tfile in .spm format containing the reference network from which is extracted the number of nodes of the simulated network\n";
  print "\t--undirected\tspecifies that the simulated network is undirected\n";
  print "\t--loop\t\tenables loop in the simulated network\n";
  exit(0);
}


open MODEL,"$ARGV[0]" or die "File \"$ARGV[0]\" opening problem\n";
# Q
my $Q;
my $l=<MODEL>;
$l=<MODEL>;
if ($l=~/^(\d+)$/){
  $Q=$1;
}
else{
  die "File \"$ARGV[0]\" format problem for # Q\n"
}
# Alpha
$l=<MODEL>;
$l=<MODEL>;
chomp($l);
my @al = split /\s+/, $l;
if (($#al+1)!=$Q){
  die "File \"$ARGV[0]\" format problem for # Alpha\n"
}
# Pi
$l=<MODEL>;
$l=<MODEL>;
my @pi=();
while ($l=~/^\d+/){
  chomp($l);
  my @li=split /\s+/, $l;
  if (($#li+1)!=$Q){
    die "File \"$ARGV[0]\" format problem for # Pi\n";
  }
  push @pi, @li;
  $l=<MODEL>;
}
close(MODEL);

# class existing
open CLASSES,"$ARGV[1]" or die "File \"$ARGV[1]\" opening problem\n";
my @classes=();
$l=<CLASSES>;
while ( defined($l) ) {
    if ($l!~/^#/){ 
	my @tmp = split /\s+/, $l;
	push @classes, $tmp[1];
    }
    $l=<CLASSES>;
}

# class simulation
my @classesim=();
my @proport=(0)x$Q;
foreach my $i(1..$nbnodes){
  my $ra=rand();
  my $sum=0;
  my $q=0;
  while ($q<$Q){
    $sum += $al[$q];
    if ( $ra<$sum ){
      push @classesim, $q;
      $proport[$q]++;
      $q=$Q;
    }
    else{
      $q++;
    }
  }
  if (($#classesim+1)<$i){
    push @classesim, $Q-1;
  }
}
#@classesim = sort {$a <=> $b} @classesim;


# summary
my $nboldnodes = $#classes+1;
my $tnbnodes = $nbnodes + $nboldnodes;
my @tclasses = ();
push @tclasses, @classes;
push @tclasses, @classesim;

if ($verbose){
  foreach my $i(0..$#classesim){print STDERR ($nboldnodes+$i),"\t$classesim[$i]\n";}
}

# edge simulation
if ($loop){
    
    if ($undirected){
	for my $i(0..($tnbnodes-1)){
	    for my $j($i..($tnbnodes-1)){
		my $ra=rand();
		if ($ra<=$pi[$tclasses[$i]*$Q+$tclasses[$j]]){
		    if (($i>=$nboldnodes)or($j>=$nboldnodes)){	
			print "$i\t$j\n";	
			print "$j\t$i\n";	
		    } 
		}	  
	    }
	}
    }
    else{
	for my $i(0..($tnbnodes-1)){
	    for my $j(0..($tnbnodes-1)){
		my $ra=rand();
		if ($ra<=$pi[$tclasses[$i]*$Q+$tclasses[$j]]){
		    if ($i != $j){
			if (($i>=$nboldnodes)or($j>=$nboldnodes)){
			    print "$i\t$j\n";
			}
		    }
		}
	    }
	}
    }
    
}
else{
    
    if ($undirected){
	for my $i(0..($tnbnodes-1)){
	    for my $j(($i+1)..($tnbnodes-1)){
		my $ra=rand();
		if ($ra<=$pi[$tclasses[$i]*$Q+$tclasses[$j]]){
		    if (($i>=$nboldnodes)or($j>=$nboldnodes)){
			print "$i\t$j\n";
			print "$j\t$i\n";
		    }
		}
	    }
	}
    }
    else{
	for my $i(0..($tnbnodes-1)){
	    for my $j(0..($i-1)){
		my $ra=rand();
		if ($ra<=$pi[$tclasses[$i]*$Q+$tclasses[$j]]){
		    if (($i>=$nboldnodes)or($j>=$nboldnodes)){
			print "$i\t$j\n";
		    }
		}
	    }
	    for my $j(($i+1)..($tnbnodes-1)){
		my $ra=rand();
		if ($ra<=$pi[$tclasses[$i]*$Q+$tclasses[$j]]){
		    if (($i>=$nboldnodes)or($j>=$nboldnodes)){
			print "$i\t$j\n";
		    }
		}
	    }
	}
    }
    
}
